# -*- coding: utf-8 -*-
'''base stuf'''

from operator import methodcaller
from itertools import chain, starmap
from collections import Mapping, Sequence, MutableMapping

from stuf.six import items, strings, keys, values
from stuf.utils import clsname, lazy, recursive_repr, exhaust, imap

__all__ = ['corestuf', 'directstuf', 'wrapstuf', 'writestuf', 'writewrapstuf']


class corestuf(object):

    '''stuf core stuff'''

    _map = dict
    _reserved = ['allowed', '_wrapped', '_map']

    def __getattr__(self, key, _getter=object.__getattribute__):
        try:
            return self[key]
        except KeyError:
            if key == 'iteritems':
                return items(self)
            elif key == 'iterkeys':
                return keys(self)
            elif key == 'itervalues':
                return values(self)
            return _getter(self, key)

    @recursive_repr
    def __repr__(self, _clsname=clsname, _mcaller=methodcaller):
        if not self:
            return '%s()' % _clsname(self)
        return '%s(%r)' % (_clsname(self), _mcaller('items')(self))

    @lazy
    def _classkeys(self):
        # protected keywords
        return frozenset(chain(
            keys(vars(self)), keys(vars(self.__class__)), self._reserved,
        ))

    @classmethod
    def _build(cls, iterable, _map=imap, _is=isinstance, _list=exhaust):
        kind = cls._map
        # add class to handle potential nested objects of the same class
        kw = kind()
        update = kw.update
        if _is(iterable, Mapping):
            update(kind(items(iterable)))
        elif _is(iterable, Sequence):
            # extract appropriate key-values from sequence
            def _coro(arg, update=update):
                try:
                    update(arg)
                except (ValueError, TypeError):
                    pass
            _list(_map(_coro, iterable))
        return kw

    @classmethod
    def _mapping(cls, iterable):
        return cls._map(iterable)

    @classmethod
    def _new(cls, iterable):
        return cls(cls._build(iterable))

    @classmethod
    def _populate(cls, past, future, _is=isinstance, m=starmap, _i=items):
        def _coro(key, value, new=cls._new, _is=_is):
            if _is(value, (Sequence, Mapping)) and not _is(value, strings):
                # see if stuf can be converted to nested stuf
                trial = new(value)
                if len(trial) > 0:
                    future[key] = trial
                else:
                    future[key] = value
            else:
                future[key] = value
        exhaust(m(_coro, _i(past)))
        return cls._postpopulate(future)

    @classmethod
    def _postpopulate(cls, future):
        return future

    def _prepopulate(self, *args, **kw):
        kw.update(self._build(args))
        return kw

    def copy(self):
        return self._build(self._map(self))


class writestuf(corestuf):

    '''stuf basestuf'''

    def __setattr__(self, key, value):
        # handle normal object attributes
        if key == '_classkeys' or key in self._classkeys:
            self.__dict__[key] = value
        # handle special attributes
        else:
            try:
                self[key] = value
            except:
                raise AttributeError(key)

    def __delattr__(self, key):
        # allow deletion of key-value pairs only
        if not key == '_classkeys' or key in self._classkeys:
            try:
                del self[key]
            except KeyError:
                raise AttributeError(key)

    def __iter__(self):
        cls = self.__class__
        for key, value in items(self):
            # nested stuf of some sort
            if isinstance(value, cls):
                yield key, dict(iter(value))
            # normal key, value pair
            else:
                yield key, value

    def __getstate__(self):
        return self._mapping(self)

    def __setstate__(self, state):
        return self._build(state)

    def update(self, *args, **kw):
        self._populate(self._prepopulate(*args, **kw), self)


class directstuf(writestuf):

    '''stuf basestuf'''

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        self.update(*args, **kw)


class wrapstuf(corestuf):

    def __init__(self, *args, **kw):
        '''
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        super(wrapstuf, self).__init__()
        self._wrapped = self._populate(
            self._prepopulate(*args, **kw), self._map(),
        )

    @classmethod
    def _postpopulate(cls, future):
        return cls._mapping(future)


class writewrapstuf(wrapstuf, writestuf, MutableMapping):

    '''wraps mappings for stuf compliance'''

    def __getitem__(self, key):
        return self._wrapped[key]

    def __setitem__(self, key, value):
        self._wrapped[key] = value

    def __delitem__(self, key):
        del self._wrapped[key]

    def __iter__(self):
        return iter(self._wrapped)

    def __len__(self):
        return len(self._wrapped)

    def clear(self):
        self._wrapped.clear()



# -*- coding: utf-8 -*-
'''stuf utilities'''

from inspect import ismodule
from itertools import starmap
try:
    from collections import OrderedDict
except  ImportError:
    from ordereddict import OrderedDict
try:
    from thread import get_ident
except ImportError:
    try:
        from dummy_thread import get_ident
    except ImportError:
        from _thread import get_ident
from functools import wraps, update_wrapper
from operator import itemgetter, attrgetter, getitem

from stuf.six import items
from stuf.six.moves import filter, map  # @UnresolvedImport

imap = map
ifilter = filter


def attr_or_item(this, key):
    '''
    get attribute or item

    @param this: object
    @param key: key to lookup
    '''
    try:
        return getitem(this, key)
    except KeyError:
        return getter(this, key)


def clsname(this):
    '''
    get class name

    @param this: object
    '''
    return getattr(this.__class__, '__name__')


def deepget(this, key, default=None):
    '''
    get an attribute with deep attribute path

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return attrgetter(key)(this)
    except AttributeError:
        return default


def exhaust(iterable, exception=StopIteration, _n=next):
    '''
    call next on an iterator until it's exhausted

    @param iterable: an iterable to exhaust
    @param exception: exception that marks end of iteration
    '''
    try:
        while True:
            _n(iterable)
    except exception:
        pass


def breakcount(func, length):
    '''
    run an iterator until it reaches its original length

    @param iterable: an iterable to exhaust
    '''
    while length:
        yield func()
        length -= 1


def exhaustmap(mapping, call, filter=False, exception=StopIteration, _n=next):
    '''
    call `next` on an iterator until it's exhausted

    @param mapping: a mapping to exhaust
    @param call: call to handle what survives the filter
    @param filter: a filter to apply to mapping (default: `None`)
    @param exception: exception sentinel (default: `StopIteration`)
    '''
    subiter = ifilter(filter, items(mapping)) if filter else items(mapping)
    try:
        while True:
            _n(starmap(call, subiter))
    except exception:
        pass


def deleter(this, key):
    '''
    delete an attribute

    @param this: object
    @param key: key to lookup
    '''
    try:
        object.__delattr__(this, key)
    except (TypeError, AttributeError):
        delattr(this, key)


def getcls(this):
    '''
    get class of instance

    @param this: an instance
    '''
    return getter(this, '__class__')


def getter(this, key):
    '''
    get an attribute

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return object.__getattribute__(this, key)
    except (AttributeError, TypeError):
        return getattr(this, key)


def getdefault(this, key, default=None):
    '''
    get an attribute

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return getter(this, key)
    except AttributeError:
        return default


def instance_or_class(key, this, that):
    '''
    get attribute of an instance or class

    @param key: name of attribute to look for
    @param this: instance to check for attribute
    @param that: class to check for attribute
    '''
    try:
        return getter(this, key)
    except AttributeError:
        return getter(that, key)


def inverse_lookup(value, this, default=None):
    '''
    get attribute of an instance by value

    @param value: value to lookup as a key
    @param this: instance to check for attribute
    @param default: default key (default: None)
    '''
    try:
        return itemgetter(value)(
            dict((v, k) for k, v in items(vars(this)))
        )
    except (TypeError, KeyError):
        return default


def iterexcept(func, exception):
    '''
    call a function repeatedly until an exception is raised

    Converts a call-until-exception interface to an iterator interface. Like
    `__builtin__.iter(func, sentinel)` but uses an exception instead of a
    sentinel to end the loop.
    '''
    try:
        while True:
            yield func()
    except exception:
        pass


def lru(this, maxsize=100):
    '''
    least-recently-used cache decorator from Raymond Hettinger

    arguments to the cached function must be hashable.

    @param maxsize: maximum number of results in LRU cache (default: 100)
    '''
    # order: least recent to most recent
    cache = OrderedDict()
    @wraps(this) #@IgnorePep8
    def wrapper(*args, **kw):
        key = args
        if kw:
            key += tuple(sorted(items(kw)))
        try:
            result = cache.pop(key)
        except KeyError:
            result = this(*args, **kw)
            # purge least recently used cache entry
            if len(cache) >= maxsize:
                cache.popitem(0)
        # record recent use of this key
        cache[key] = result
        return result
    return wrapper


def recursive_repr(this):
    '''
    Decorator to make a repr function return "..." for a recursive call

    @param this: object
    '''
    repr_running = set()
    def wrapper(self): #@IgnorePep8
        key = id(self), get_ident()
        if key in repr_running:
            return '...'
        repr_running.add(key)
        try:
            result = this(self)
        finally:
            repr_running.discard(key)
        return result
    # Can't use functools.wraps() here because of bootstrap issues
    wrapper.__module__ = getattr(this, '__module__')
    wrapper.__doc__ = getattr(this, '__doc__')
    wrapper.__name__ = selfname(this)
    return wrapper


def selfname(this):
    '''
    get object name

    @param this: object
    '''
    return getter(this, '__name__')


def setter(this, key, value):
    '''
    set attribute

    @param this: object
    @param key: key to set
    @param value: value to set
    '''
    # it's an instance
    try:
        this.__dict__[key] = value
        return value
    # it's a class
    except TypeError:
        setattr(this, key, value)
        return value


def setdefault(this, key, default=None):
    '''
    get an attribute, creating and setting it if needed

    @param this: object
    @param key: key to lookup
    @param default: default value returned if key not found (default: None)
    '''
    try:
        return getter(this, key)
    except AttributeError:
        setter(this, key, default)
        return default


class lazybase(object):

    '''base for lazy descriptors'''


class _lazyinit(lazybase):

    '''base for lazy descriptors'''

    def __init__(self, method, _wrap=update_wrapper):
        super(_lazyinit, self).__init__()
        self.method = method
        self.name = selfname(method)
        _wrap(self, method)

    def _set(self, this):
        return setter(this, self.name, self.method(this))


class lazy(_lazyinit):

    '''lazily assign attributes on an instance upon first use.'''

    def __get__(self, this, that):
        return self if this is None else self._set(this)


class lazy_class(_lazyinit):

    '''lazily assign attributes on an class upon first use.'''

    def __get__(self, this, that):
        return self._set(that)


class lazy_set(lazy):

    '''lazy assign attributes with a custom setter'''

    def __init__(self, method, fget=None, _wrap=update_wrapper):
        super(lazy_set, self).__init__(method)
        self.fget = fget
        _wrap(self, method)

    def __set__(self, this, value):
        self.fget(this, value)

    def __delete__(self, this):
        del this.__dict__[self.name]

    def setter(self, func):
        self.fget = func
        return self


class bi(_lazyinit):

    '''call as both class and instance method'''

    def __get__(self, this, that):
        return self._factory(that) if this is None else self._factory(this)

    def _factory(self, this):
        def func(*args, **kw):
            return self.method(*(this,) + args, **kw)
        setattr(this, self.name, func)
        return func


class bothbase(_lazyinit):

    def __init__(self, method, expr=None, _wrap=update_wrapper):
        super(bothbase, self).__init__(method)
        self.expr = expr or method
        _wrap(self, method)

    def expression(self, expr):
        '''modifying decorator that defines a general method'''
        self.expr = expr
        return self


class both(bothbase):

    '''
    descriptor that caches results of instance-level results while allowing
    class-level results
    '''

    def __get__(self, this, that):
        return self.expr(that) if this is None else self._set(this)


class either(bothbase):

    '''
    descriptor that caches results of both instance- and class-level results
    '''

    def __get__(self, this, that):
        if this is None:
            return setter(that, self.name, self.expr(that))
        return self._set(this)


class twoway(bothbase):

    '''descriptor that enables instance and class-level results'''

    def __get__(self, this, that):
        return self.expr(that) if this is None else self.method(this)


lru_wrapped = lru
get_or_default = getdefault

__all__ = sorted(name for name, obj in items(locals()) if not any([
    name.startswith('_'), ismodule(obj),
]))
del ismodule



# -*- coding: utf-8 -*-
# pylint: disable-msg=w0108
'''test stuf'''

try:
    import unittest2 as unittest
except ImportError:
    import unittest


class TestStuf(unittest.TestCase):

    @property
    def _makeone(self):
        from stuf import stuf
        return stuf

    @property
    def _maketwo(self):
        from stuf import istuf
        return istuf

    def setUp(self):
        self.stuf = self._makeone(
            test1='test1', test2='test2', test3=dict(e=1)
        )

    def test__getattr__(self):
        self.assertEqual(self.stuf.test1, 'test1')
        self.assertEqual(self.stuf.test2, 'test2')
        self.assertEqual(self.stuf.test3.e, 1)

    def test__getitem__(self):
        self.assertEqual(self.stuf['test1'], 'test1')
        self.assertEqual(self.stuf['test2'], 'test2')
        self.assertEqual(self.stuf['test3']['e'], 1)

    def test__setattr__(self):
        self.stuf.max = 3
        self.stuf.test1 = 'test1again'
        self.stuf.test2 = 'test2again'
        self.stuf.test3.e = 5
        self.assertEqual(self.stuf.max, 3)
        self.assertEqual(self.stuf.test1, 'test1again')
        self.assertEqual(self.stuf.test2, 'test2again')
        self.assertEqual(self.stuf.test3.e, 5)

    def test__setitem__(self):
        self.stuf['max'] = 3
        self.stuf['test1'] = 'test1again'
        self.stuf['test2'] = 'test2again'
        self.stuf['test3']['e'] = 5
        self.assertEqual(self.stuf['max'], 3)
        self.assertEqual(self.stuf['test1'], 'test1again')
        self.assertEqual(self.stuf['test2'], 'test2again')
        self.assertEqual(self.stuf['test3']['e'], 5)

    def test__delattr__(self):
        del self.stuf.test1
        del self.stuf.test2
        del self.stuf.test3.e
        self.assertTrue(len(self.stuf.test3) == 0)
        del self.stuf.test3
        self.assertTrue(len(self.stuf) == 0)
        self.assertRaises(AttributeError, lambda: self.stuf.test1)
        self.assertRaises(AttributeError, lambda: self.stuf.test2)
        self.assertRaises(AttributeError, lambda: self.stuf.test3)
        self.assertRaises(AttributeError, lambda: self.stuf.test3.e)

    def test__delitem__(self):
        del self.stuf['test1']
        del self.stuf['test2']
        del self.stuf['test3']['e']
        self.assertFalse('e' in self.stuf['test3'])
        self.assertTrue(len(self.stuf['test3']) == 0)
        del self.stuf['test3']
        self.assertTrue(len(self.stuf) == 0)
        self.assertFalse('test1' in self.stuf)
        self.assertFalse('test2' in self.stuf)
        self.assertFalse('test3' in self.stuf)

    def test_get(self):
        self.assertEqual(self.stuf.get('test1'), 'test1')
        self.assertEqual(self.stuf.get('test2'), 'test2')
        self.assertIsNone(self.stuf.get('test4'), 'test4')
        self.assertEqual(self.stuf.get('test3').get('e'), 1)
        self.assertIsNone(self.stuf.get('test3').get('r'))

    def test__cmp__(self):
        tstuff = self._maketwo(
            (('test1', 'test1'), ('test2', 'test2'), ('test3', dict(e=1)))
        )
        self.assertEqual(self.stuf, tstuff)

    def test__len__(self):
        self.assertEqual(len(self.stuf), 3)
        self.assertEqual(len(self.stuf.test3), 1)

    def test_clear(self):
        self.stuf.test3.clear()
        self.assertEqual(len(self.stuf.test3), 0)
        self.stuf.clear()
        self.assertEqual(len(self.stuf), 0)

    def test_items(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', {'e': 1}) in slist)

    def test_iteritems(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', {'e': 1}) in slist)

    def test_iterkeys(self):
        slist = list(self.stuf.keys())
        slist2 = list(self.stuf.test3.keys())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue('test3' in slist)
        self.assertTrue('e' in slist2)

    def test_itervalues(self):
        slist = list(self.stuf.values())
        slist2 = list(self.stuf.test3.values())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue({'e': 1} in slist)
        self.assertTrue(1 in slist2)

    def test_pop(self):
        self.assertEqual(self.stuf.test3.pop('e'), 1)
        self.assertEqual(self.stuf.pop('test1'), 'test1')
        self.assertEqual(self.stuf.pop('test2'), 'test2')
        self.assertEqual(self.stuf.pop('test3'), {})

    def test_popitem(self):
        item = self.stuf.popitem()
        self.assertEqual(len(item) + len(self.stuf), 4, item)

    def test_setdefault(self):
        self.assertEqual(self.stuf.test3.setdefault('e', 8), 1)
        self.assertEqual(self.stuf.test3.setdefault('r', 8), 8)
        self.assertEqual(self.stuf.setdefault('test1', 8), 'test1')
        self.assertEqual(self.stuf.setdefault('pow', 8), 8)

    def test_update(self):
        tstuff = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        tstuff['test1'] = 3
        tstuff['test2'] = 6
        tstuff['test3'] = dict(f=2)
        self.stuf.update(tstuff)
        self.assertEqual(self.stuf['test1'], 3, self.stuf.items())
        self.assertEqual(self.stuf['test2'], 6)
        self.assertEqual(self.stuf['test3'], dict(f=2), self.stuf)

    def test_values(self):
        slist1 = self.stuf.test3.values()
        slist2 = self.stuf.values()
        self.assertTrue(1 in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue({'e': 1} in slist2)

    def test_keys(self):
        slist1 = self.stuf.test3.keys()
        slist2 = self.stuf.keys()
        self.assertTrue('e' in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test3' in slist2)

    def test_pickle(self):
        import pickle
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        pkle = pickle.dumps(tstuf)
        nstuf = pickle.loads(pkle)
        self.assertEquals(tstuf, nstuf)

    def test_copy(self):
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        nstuf = tstuf.copy()
        self.assertEquals(tstuf, nstuf)


class TestDefaultStuf(unittest.TestCase):

    @property
    def _makeone(self):
        from stuf import defaultstuf
        return defaultstuf

    @property
    def _maketwo(self):
        from stuf import idefaultstuf
        return idefaultstuf

    def setUp(self):
        self.stuf = self._makeone(
            list, test1='test1', test2='test2', test3=dict(e=1)
        )

    def test__getattr__(self):
        self.assertEqual(self.stuf.test1, 'test1')
        self.assertEqual(self.stuf.test2, 'test2')
        self.assertEqual(self.stuf.test4, [])
        self.assertEqual(self.stuf.test3.e, 1)
        self.assertEqual(self.stuf.test3.f, [])

    def test__getitem__(self):
        self.assertEqual(self.stuf['test1'], 'test1')
        self.assertEqual(self.stuf['test2'], 'test2')
        self.assertEqual(self.stuf['test4'], [])
        self.assertEqual(self.stuf['test3']['e'], 1)
        self.assertEqual(self.stuf['test3']['f'], [])

    def test__setattr__(self):
        self.stuf.max = 3
        self.stuf.test1 = 'test1again'
        self.stuf.test2 = 'test2again'
        self.stuf.test3.e = 5
        self.assertEqual(self.stuf.max, 3)
        self.assertEqual(self.stuf.test1, 'test1again')
        self.assertEqual(self.stuf.test2, 'test2again')
        self.assertEqual(self.stuf.test3.e, 5)

    def test__setitem__(self):
        self.stuf['max'] = 3
        self.stuf['test1'] = 'test1again'
        self.stuf['test2'] = 'test2again'
        self.stuf['test3']['e'] = 5
        self.assertEqual(self.stuf['max'], 3)
        self.assertEqual(self.stuf['test1'], 'test1again')
        self.assertEqual(self.stuf['test2'], 'test2again')
        self.assertEqual(self.stuf['test3']['e'], 5)

    def test__delattr__(self):
        del self.stuf.test1
        del self.stuf.test2
        del self.stuf.test3.e
        self.assertTrue(len(self.stuf.test3) == 0)
        del self.stuf.test3
        self.assertTrue(len(self.stuf) == 0)
        self.assertEqual(self.stuf.test1, [])
        self.assertEqual(self.stuf.test2, [])
        self.assertEqual(self.stuf.test3, [])
        self.assertRaises(AttributeError, lambda: self.stuf.test3.e)

    def test__delitem__(self):
        del self.stuf['test1']
        del self.stuf['test2']
        del self.stuf['test3']['e']
        self.assertFalse('e' in self.stuf['test3'])
        self.assertTrue(len(self.stuf['test3']) == 0)
        self.assertEqual(self.stuf['test3']['e'], [])
        del self.stuf['test3']
        self.assertTrue(len(self.stuf) == 0)
        self.assertFalse('test1' in self.stuf)
        self.assertFalse('test2' in self.stuf)
        self.assertFalse('test3' in self.stuf)
        self.assertEqual(self.stuf['test1'], [])
        self.assertEqual(self.stuf['test2'], [])
        self.assertEqual(self.stuf['test3'], [])
        self.assertRaises(TypeError, lambda: self.stuf['test3']['e'])

    def test_get(self):
        self.assertEqual(self.stuf.get('test1'), 'test1')
        self.assertEqual(self.stuf.get('test2'), 'test2')
        self.assertIsNone(self.stuf.get('test4'), 'test4')
        self.assertEqual(self.stuf.get('test3').get('e'), 1)
        self.assertIsNone(self.stuf.get('test3').get('r'))

    def test__cmp__(self):
        tstuff = self._maketwo(
            list,
            (1, 2),
            {},
            (('test1', 'test1'), ('test2', 'test2'), ('test3', dict(e=1))),
        )
        self.assertEqual(self.stuf, tstuff)

    def test__len__(self):
        self.assertEqual(len(self.stuf), 3)
        self.assertEqual(len(self.stuf.test3), 1)

    def test_clear(self):
        self.stuf.test3.clear()
        self.assertEqual(len(self.stuf.test3), 0)
        self.assertEqual(self.stuf['test3']['e'], [])
        self.stuf.clear()
        self.assertEqual(len(self.stuf), 0)
        self.assertEqual(self.stuf['test1'], [])
        self.assertEqual(self.stuf['test2'], [])
        self.assertEqual(self.stuf['test3'], [])

    def test_items(self):
        self.assertEqual(self.stuf['test4'], [])
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', {'e': 1}) in slist)
        self.assertTrue(('test4', []) in slist)

    def test_iteritems(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', {'e': 1}) in slist)

    def test_iterkeys(self):
        slist = list(self.stuf.keys())
        slist2 = list(self.stuf.test3.keys())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue('test3' in slist)
        self.assertTrue('e' in slist2)

    def test_itervalues(self):
        slist = list(self.stuf.values())
        slist2 = list(self.stuf.test3.values())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue({'e': 1} in slist)
        self.assertTrue(1 in slist2)

    def test_pop(self):
        self.assertEqual(self.stuf.test3.pop('e'), 1)
        self.assertEqual(self.stuf.pop('test1'), 'test1')
        self.assertEqual(self.stuf.pop('test2'), 'test2')
        self.assertEqual(self.stuf.pop('test3'), {})

    def test_popitem(self):
        item = self.stuf.popitem()
        self.assertEqual(len(item) + len(self.stuf), 4, item)

    def test_setdefault(self):
        self.assertEqual(self.stuf.test3.setdefault('e', 8), 1)
        self.assertEqual(self.stuf.test3.setdefault('r', 8), 8)
        self.assertEqual(self.stuf.setdefault('test1', 8), 'test1')
        self.assertEqual(self.stuf.setdefault('pow', 8), 8)

    def test_update(self):
        tstuff = self._makeone(
            list, (1, 2), test1='test1', test2='test2', test3=dict(e=1)
        )
        tstuff['test1'] = 3
        tstuff['test2'] = 6
        tstuff['test3'] = dict(f=2)
        self.stuf.update(tstuff)
        self.assertEqual(self.stuf['test1'], 3)
        self.assertEqual(self.stuf['test2'], 6)
        self.assertEqual(self.stuf['test3'], dict(f=2), self.stuf)

    def test_values(self):
        slist1 = self.stuf.test3.values()
        slist2 = self.stuf.values()
        self.assertTrue(1 in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue({'e': 1} in slist2)

    def test_keys(self):
        slist1 = self.stuf.test3.keys()
        slist2 = self.stuf.keys()
        self.assertTrue('e' in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test3' in slist2)

    def test_pickle(self):
        import pickle
        tstuf = self._makeone(
            list, ([],), test1='test1', test2='test2', test3=dict(e=1)
        )
        pkle = pickle.dumps(tstuf)
        nstuf = pickle.loads(pkle)
        self.assertEquals(tstuf, nstuf)

    def test_copy(self):
        tstuf = self._makeone(
            list, ([],), test1='test1', test2='test2', test3=dict(e=1)
        )
        nstuf = tstuf.copy()
        self.assertEquals(tstuf, nstuf)


class TestFixedStuf(unittest.TestCase):

    @property
    def _makeone(self):
        from stuf import fixedstuf
        return fixedstuf

    @property
    def _maketwo(self):
        from stuf import ifixedstuf
        return ifixedstuf

    def setUp(self):
        self.stuf = self._makeone(
            test1='test1', test2='test2', test3=dict(e=1)
        )

    def test__getattr__(self):
        self.assertEqual(self.stuf.test1, 'test1')
        self.assertEqual(self.stuf.test2, 'test2')
        self.assertEqual(self.stuf.test3.e, 1)

    def test__getitem__(self):
        self.assertEqual(self.stuf['test1'], 'test1')
        self.assertEqual(self.stuf['test2'], 'test2')
        self.assertEqual(self.stuf['test3']['e'], 1)

    def test__setattr__(self):
        self.assertRaises(AttributeError, lambda: setattr(self.stuf, 'max', 3))
        self.stuf.test1 = 'test1again'
        self.stuf.test2 = 'test2again'
        self.stuf.test3.e = 5
        self.assertRaises(AttributeError, lambda: self.stuf.max)
        self.assertEqual(self.stuf.test1, 'test1again')
        self.assertEqual(self.stuf.test2, 'test2again')
        self.assertEqual(self.stuf.test3.e, 5)

    def test__setitem__(self):
        self.assertRaises(KeyError, lambda: self.stuf.__setitem__('max', 3))
        self.stuf['test1'] = 'test1again'
        self.stuf['test2'] = 'test2again'
        self.stuf['test3']['e'] = 5
        self.assertRaises(KeyError, lambda: self.stuf.__getitem__('max'))
        self.assertEqual(self.stuf['test1'], 'test1again')
        self.assertEqual(self.stuf['test2'], 'test2again')
        self.assertEqual(self.stuf['test3']['e'], 5)

    def test__delattr__(self):
        self.assertRaises(TypeError, lambda: delattr(self.stuf.test1))
        self.assertRaises(TypeError, lambda: delattr(self.stuf.test3.e))

    def test__delitem__(self):
        del self.stuf.test3['e']
        self.assertRaises(KeyError, lambda: self.stuf.test3['e'])
        del self.stuf['test1']
        self.assertRaises(KeyError, lambda: self.stuf['test1'])

    def test_get(self):
        self.assertEqual(self.stuf.get('test1'), 'test1')
        self.assertEqual(self.stuf.get('test2'), 'test2')
        self.assertIsNone(self.stuf.get('test4'), 'test4')
        self.assertEqual(self.stuf.get('test3').get('e'), 1)
        self.assertIsNone(self.stuf.get('test3').get('r'))

    def test__cmp__(self):
        tstuff = self._maketwo(
            (('test1', 'test1'), ('test2', 'test2'), ('test3', dict(e=1)))
        )
        self.assertEqual(self.stuf, tstuff)

    def test__len__(self):
        self.assertEqual(len(self.stuf), 3)
        self.assertEqual(len(self.stuf.test3), 1)

    def test_clear(self):
        self.assertRaises(KeyError, lambda: self.stuf.__setitem__('max', 3))
        self.stuf.clear()
        self.stuf['test1'] = 'test1again'
        self.stuf['test3'] = 5

    def test_items(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        dummy = self._maketwo({'e': 1})
        self.assertTrue(('test3', dummy) in slist, slist)

    def test_iteritems(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', self._maketwo({'e': 1})) in slist, slist)

    def test_iterkeys(self):
        slist = list(self.stuf.keys())
        slist2 = list(self.stuf.test3.keys())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue('test3' in slist)
        self.assertTrue('e' in slist2)

    def test_itervalues(self):
        slist1 = list(i for i in self.stuf.values())
        slist2 = list(i for i in self.stuf.test3.values())
        self.assertTrue(1 in slist2)
        self.assertTrue('test1' in slist1)
        self.assertTrue(self._maketwo({'e': 1}) in slist1)
        self.assertTrue('test2' in slist1)

    def test_pop(self):
        self.assertRaises(AttributeError, lambda: self.stuf.test3.pop('e'))
        self.assertRaises(AttributeError, lambda: self.stuf.pop('test1'))

    def test_popitem(self):
        self.assertRaises(AttributeError, lambda: self.stuf.test3.popitem())
        self.assertRaises(AttributeError, lambda: self.stuf.popitem())

    def test_setdefault(self):
        self.assertEqual(self.stuf.test3.setdefault('e', 8), 1)
        self.assertRaises(KeyError, lambda: self.stuf.test3.setdefault('r', 8))
        self.assertEqual(self.stuf.setdefault('test1', 8), 'test1')
        self.assertRaises(KeyError, lambda: self.stuf.setdefault('pow', 8))

    def test_update(self):
        tstuff = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        tstuff['test1'] = 3
        tstuff['test2'] = 6
        tstuff['test3'] = dict(f=2)
        self.stuf.update(tstuff)
        self.assertEqual(self.stuf['test1'], 3)
        self.assertEqual(self.stuf['test2'], 6)
        self.assertEqual(
            self.stuf['test3'], self._maketwo(dict(f=2)), self.stuf['test3']
        )

    def test_values(self):
        slist1 = self.stuf.test3.values()
        slist2 = self.stuf.values()
        self.assertTrue(1 in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue(self._maketwo({'e': 1}) in slist2)

    def test_keys(self):
        slist1 = self.stuf.test3.keys()
        slist2 = self.stuf.keys()
        self.assertTrue('e' in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test3' in slist2)

    def test_copy(self):
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        nstuf = tstuf.copy()
        self.assertEquals(tstuf, nstuf)

    def test_pickle(self):
        import pickle
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        pkle = pickle.dumps(tstuf)
        nstuf = pickle.loads(pkle)
        self.assertEquals(tstuf, nstuf)


class TestFrozenStuf(unittest.TestCase):

    @property
    def _makeone(self):
        from stuf import frozenstuf
        return frozenstuf

    @property
    def _maketwo(self):
        from stuf import ifrozenstuf
        return ifrozenstuf

    def setUp(self):
        self.stuf = self._makeone(
            test1='test1', test2='test2', test3=dict(e=1)
        )

    def test__getattr__(self):
        self.assertEqual(self.stuf.test1, 'test1')
        self.assertEqual(self.stuf.test2, 'test2')
        self.assertEqual(self.stuf.test3.e, 1)

    def test__getitem__(self):
        self.assertEqual(self.stuf['test1'], 'test1')
        self.assertEqual(self.stuf['test2'], 'test2')
        self.assertEqual(self.stuf['test3']['e'], 1)

    def test__setattr__(self):
        self.assertRaises(AttributeError, setattr(self.stuf, 'max', 3))
        self.assertRaises(
            AttributeError, setattr(self.stuf, 'test1', 'test1again')
        )
        self.assertRaises(
            AttributeError, setattr(self.stuf.test3, 'e', 5)
        )

    def test__setitem__(self):
        self.assertRaises(
            AttributeError, lambda: self.stuf.__setitem__('max', 3)
        )
        self.assertRaises(
            AttributeError,
            lambda: self.stuf.__setitem__('test2', 'test2again'),
        )
        self.assertRaises(
            AttributeError, lambda: self.stuf.test3.__setitem__('e', 5)
        )

    def test__delattr__(self):
        self.assertRaises(TypeError, lambda: delattr(self.stuf.test1))
        self.assertRaises(TypeError, lambda: delattr(self.stuf.test3.e))

    def test__delitem__(self):
        self.assertRaises(
            AttributeError, lambda: self.stuf.__delitem__('test1'),
        )
        self.assertRaises(
            AttributeError, lambda: self.stuf.test3.__delitem__('test1'),
        )

    def test_get(self):
        self.assertEqual(self.stuf.get('test1'), 'test1')
        self.assertEqual(self.stuf.get('test2'), 'test2')
        self.assertIsNone(self.stuf.get('test4'), 'test4')
        self.assertEqual(self.stuf.get('test3').get('e'), 1)
        self.assertIsNone(self.stuf.get('test3').get('r'))

    def test__cmp__(self):
        tstuff = self._maketwo(
            (('test1', 'test1'), ('test2', 'test2'), ('test3', dict(e=1)))
        )
        self.assertEqual(self.stuf, tstuff)

    def test__len__(self):
        self.assertEqual(len(self.stuf), 3)
        self.assertEqual(len(self.stuf.test3), 1)

    def test_clear(self):
        self.assertRaises(AttributeError, lambda: self.stuf.test3.clear())
        self.assertRaises(AttributeError, lambda: self.stuf.clear())

    def test_items(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', self._maketwo({'e': 1})) in slist, slist)

    def test_iteritems(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', self._maketwo({'e': 1})) in slist, slist)

    def test_iterkeys(self):
        slist = list(self.stuf.keys())
        slist2 = list(self.stuf.test3.keys())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue('test3' in slist)
        self.assertTrue('e' in slist2)

    def test_itervalues(self):
        slist1 = list(self.stuf.values())
        slist2 = list(self.stuf.test3.values())
        self.assertTrue('test2' in slist1)
        self.assertTrue(self._maketwo({'e': 1}) in slist1)
        self.assertTrue(1 in slist2)
        self.assertTrue('test1' in slist1)

    def test_pop(self):
        self.assertRaises(AttributeError, lambda: self.stuf.test3.pop('e'))
        self.assertRaises(AttributeError, lambda: self.stuf.pop('test1'))

    def test_popitem(self):
        self.assertRaises(AttributeError, lambda: self.stuf.test3.popitem())
        self.assertRaises(AttributeError, lambda: self.stuf.popitem())

    def test_setdefault(self):
        self.assertRaises(
            AttributeError, lambda: self.stuf.test3.setdefault('e', 8)
        )
        self.assertRaises(
            AttributeError, lambda: self.stuf.test3.setdefault('r', 8)
        )
        self.assertRaises(
            AttributeError, lambda: self.stuf.setdefault('test1', 8)
        )
        self.assertRaises(
            AttributeError, lambda: self.stuf.setdefault('pow', 8)
        )

    def test_update(self):
        tstuff = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        self.assertRaises(
            AttributeError, lambda: self.stuf.test3.update(tstuff),
        )
        self.assertRaises(AttributeError, lambda: self.stuf.update(tstuff))

    def test_values(self):
        slist1 = self.stuf.test3.values()
        slist2 = self.stuf.values()
        self.assertTrue(1 in slist1)
        self.assertTrue(self._maketwo({'e': 1}) in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test1' in slist2)

    def test_keys(self):
        slist1 = self.stuf.test3.keys()
        slist2 = self.stuf.keys()
        self.assertTrue('e' in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test3' in slist2)

    def test_copy(self):
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        nstuf = tstuf.copy()
        self.assertEquals(tstuf, nstuf)

    def test_pickle(self):
        import pickle
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        pkle = pickle.dumps(tstuf)
        nstuf = pickle.loads(pkle)
        self.assertEquals(tstuf, nstuf)


class TestOrderedStuf(unittest.TestCase):

    @property
    def _makeone(self):
        from stuf import orderedstuf
        return orderedstuf

    @property
    def _maketwo(self):
        from stuf import iorderedstuf
        return iorderedstuf

    def setUp(self):
        self.stuf = self._makeone(
            test1='test1', test2='test2', test3=dict(e=1)
        )

    def test__getattr__(self):
        self.assertEqual(self.stuf.test1, 'test1')
        self.assertEqual(self.stuf.test2, 'test2')
        self.assertEqual(self.stuf.test3.e, 1)

    def test__getitem__(self):
        self.assertEqual(self.stuf['test1'], 'test1')
        self.assertEqual(self.stuf['test2'], 'test2')
        self.assertEqual(self.stuf['test3']['e'], 1)

    def test__setattr__(self):
        self.stuf.max = 3
        self.stuf.test1 = 'test1again'
        self.stuf.test2 = 'test2again'
        self.stuf.test3.e = 5
        self.assertEqual(self.stuf.max, 3)
        self.assertEqual(self.stuf.test1, 'test1again')
        self.assertEqual(self.stuf.test2, 'test2again')
        self.assertEqual(self.stuf.test3.e, 5)

    def test__setitem__(self):
        self.stuf['max'] = 3
        self.stuf['test1'] = 'test1again'
        self.stuf['test2'] = 'test2again'
        self.stuf['test3']['e'] = 5
        self.assertEqual(self.stuf['max'], 3)
        self.assertEqual(self.stuf['test1'], 'test1again')
        self.assertEqual(self.stuf['test2'], 'test2again')
        self.assertEqual(self.stuf['test3']['e'], 5)

    def test__delattr__(self):
        del self.stuf.test1
        del self.stuf.test2
        del self.stuf.test3.e
        self.assertTrue(len(self.stuf.test3) == 0)
        del self.stuf.test3
        self.assertTrue(len(self.stuf) == 0)
        self.assertRaises(AttributeError, lambda: self.stuf.test1)
        self.assertRaises(AttributeError, lambda: self.stuf.test2)
        self.assertRaises(AttributeError, lambda: self.stuf.test3)
        self.assertRaises(AttributeError, lambda: self.stuf.test3.e)

    def test__delitem__(self):
        del self.stuf['test1']
        del self.stuf['test2']
        del self.stuf['test3']['e']
        self.assertFalse('e' in self.stuf['test3'])
        self.assertTrue(len(self.stuf['test3']) == 0)
        del self.stuf['test3']
        self.assertTrue(len(self.stuf) == 0)
        self.assertFalse('test1' in self.stuf)
        self.assertFalse('test2' in self.stuf)
        self.assertFalse('test3' in self.stuf)

    def test_get(self):
        self.assertEqual(self.stuf.get('test1'), 'test1')
        self.assertEqual(self.stuf.get('test2'), 'test2')
        self.assertIsNone(self.stuf.get('test4'), 'test4')
        self.assertEqual(self.stuf.get('test3').get('e'), 1)
        self.assertIsNone(self.stuf.get('test3').get('r'))

    def test__cmp__(self):
        tstuff = self._maketwo(
            (('test1', 'test1'), ('test2', 'test2'), ('test3', dict(e=1)))
        )
        self.assertEqual(self.stuf, tstuff)

    def test__len__(self):
        self.assertEqual(len(self.stuf), 3)
        self.assertEqual(len(self.stuf.test3), 1)

    def test_clear(self):
        self.stuf.test3.clear()
        self.assertEqual(len(self.stuf.test3), 0)
        self.stuf.clear()
        self.assertEqual(len(self.stuf), 0)

    def test_items(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', {'e': 1}) in slist)

    def test_iteritems(self):
        slist = list(self.stuf.items())
        self.assertTrue(('test1', 'test1') in slist)
        self.assertTrue(('test2', 'test2') in slist)
        self.assertTrue(('test3', self._maketwo({'e': 1})) in slist)

    def test_iterkeys(self):
        slist = list(self.stuf.keys())
        slist2 = list(self.stuf.test3.keys())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue('test3' in slist)
        self.assertTrue('e' in slist2)

    def test_itervalues(self):
        slist = list(self.stuf.values())
        slist2 = list(self.stuf.test3.values())
        self.assertTrue('test1' in slist)
        self.assertTrue('test2' in slist)
        self.assertTrue(self._maketwo({'e': 1}) in slist)
        self.assertTrue(1 in slist2)

    def test_pop(self):
        self.assertEqual(self.stuf.test3.pop('e'), 1)
        self.assertEqual(self.stuf.pop('test1'), 'test1')
        self.assertEqual(self.stuf.pop('test2'), 'test2')
        self.assertEqual(self.stuf.pop('test3'), {})

    def test_popitem(self):
        item = self.stuf.popitem()
        self.assertEqual(len(item) + len(self.stuf), 4, item)

    def test_setdefault(self):
        self.assertEqual(self.stuf.test3.setdefault('e', 8), 1)
        self.assertEqual(self.stuf.test3.setdefault('r', 8), 8)
        self.assertEqual(self.stuf.setdefault('test1', 8), 'test1')
        self.assertEqual(self.stuf.setdefault('pow', 8), 8)

    def test_update(self):
        tstuff = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        tstuff['test1'] = 3
        tstuff['test2'] = 6
        tstuff['test3'] = dict(f=2)
        self.stuf.update(tstuff)
        self.assertEqual(self.stuf['test1'], 3)
        self.assertEqual(self.stuf['test2'], 6)
        self.assertEqual(self.stuf['test3'], dict(f=2), self.stuf)

    def test_values(self):
        slist1 = self.stuf.test3.values()
        slist2 = self.stuf.values()
        self.assertTrue(1 in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue(self._maketwo({'e': 1}) in slist2)

    def test_keys(self):
        slist1 = self.stuf.test3.keys()
        slist2 = self.stuf.keys()
        self.assertTrue('e' in slist1)
        self.assertTrue('test1' in slist2)
        self.assertTrue('test2' in slist2)
        self.assertTrue('test3' in slist2)

    def test_copy(self):
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        nstuf = tstuf.copy()
        self.assertEquals(tstuf, nstuf)

    def test_pickle(self):
        import pickle
        tstuf = self._makeone(test1='test1', test2='test2', test3=dict(e=1))
        pkle = pickle.dumps(tstuf)
        nstuf = pickle.loads(pkle)
        self.assertEquals(tstuf, nstuf)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
'''stuf has attributes'''

from stuf.core import defaultstuf, fixedstuf, frozenstuf, orderedstuf, stuf

__all__ = ('defaultstuf', 'fixedstuf', 'frozenstuf', 'orderedstuf', 'stuf')
__version__ = (0, 8, 9)


idefaultstuf = defaultstuf
ifixedstuf = fixedstuf
ifrozenstuf = frozenstuf
iorderedstuf = orderedstuf
istuf = stuf



# -*- coding: utf-8 -*-
'''core stuf'''

from itertools import starmap
from collections import Mapping, Sequence, defaultdict, namedtuple

from stuf.six import items, keys, values
from stuf.utils import OrderedDict, getter, imap, exhaust
from stuf.base import directstuf, wrapstuf, writewrapstuf

__all__ = ('defaultstuf', 'fixedstuf', 'frozenstuf', 'orderedstuf', 'stuf')


class defaultstuf(directstuf, defaultdict):

    '''
    dictionary with attribute-style access and a factory function to provide a
    default value for keys with no value
    '''

    _map = defaultdict

    def __getattr__(self, key, _getter=object.__getattribute__):
        try:
            if key == 'iteritems':
                return items(self)
            elif key == 'iterkeys':
                return keys(self)
            elif key == 'itervalues':
                return values(self)
            return _getter(self, key)
        except AttributeError:
            return self[key]

    def __init__(self, default, *args, **kw):
        '''
        @param default: function that can provide default values
        @param *args: iterable of keys/value pairs
        @param **kw: keyword arguments
        '''
        defaultdict.__init__(self, default)
        directstuf.__init__(self, *args, **kw)

    @classmethod
    def _build(cls, default, iterable, _map=imap, _list=exhaust):
        kind = cls._map
        # add class to handle potential nested objects of the same class
        kw = kind(default)
        update = kw.update
        if isinstance(iterable, Mapping):
            update(kind(default, iterable))
        elif isinstance(iterable, Sequence):
            # extract appropriate key-values from sequence
            def _coro(arg):
                try:
                    update(arg)
                except (ValueError, TypeError):
                    pass
            _list(_map(_coro, iterable))
        return kw

    @classmethod
    def _new(cls, default, iterable):
        return cls(default, cls._build(default, iterable))

    def _populate(self, past, future, _map=starmap, _items=items):
        def _coro(key, value, new=self._new):
            if isinstance(value, (Mapping, Sequence)):
                # see if stuf can be converted to nested stuf
                trial = new(self.default_factory, value)
                if len(trial) > 0:
                    future[key] = trial
                else:
                    future[key] = value
            else:
                future[key] = value
        exhaust(_map(_coro, _items(past)))

    def _prepopulate(self, *args, **kw):
        kw.update(self._build(self.default_factory, args))
        return kw

    def copy(self):
        return self._build(self.default_factory, dict(self))


class fixedstuf(writewrapstuf):

    '''
    dictionary with attribute-style access with mutability restricted to
    initial keys
    '''

    def __setitem__(self, key, value):
        # only access initial keys
        if key in self.allowed:
            super(fixedstuf, self).__setitem__(key, value)
        else:
            raise KeyError('key "{0}" not allowed'.format(key))

    def __reduce__(self):
        return (self.__class__, (self._wrapped.copy(),))

    def _prepopulate(self, *args, **kw):
        iterable = super(fixedstuf, self)._prepopulate(*args, **kw)
        self.allowed = frozenset(keys(iterable))
        return iterable

    def popitem(self):
        raise AttributeError()

    def pop(self, key, default=None):
        raise AttributeError()


class frozenstuf(wrapstuf, Mapping):

    '''immutable dictionary with attribute-style access'''

    __slots__ = ['_wrapped']

    def __getitem__(self, key):
        try:
            return getattr(self._wrapped, key)
        except AttributeError:
            raise KeyError('key {0} not found'.format(key))

    def __iter__(self, _iter=iter, _getr=getter):
        return _iter(_getr(self, '_wrapped')._asdict())

    def __len__(self, _len=len, _getr=getter):
        return _len(getter(self, '_wrapped')._asdict())

    def __reduce__(self, _getter=getter):
        return (self.__class__, (_getter(self, '_wrapped')._asdict().copy(),))

    @classmethod
    def _mapping(self, mapping, _namedtuple=namedtuple, _keys=keys):
        return _namedtuple('frozenstuf', _keys(mapping))(**mapping)


class orderedstuf(writewrapstuf):

    '''dictionary with dot attributes that remembers insertion order'''

    _mapping = OrderedDict

    def __reversed__(self):
        return getter(self, '_wrapped').__reversed__()

    def __reduce__(self):
        return self._wrapped.__reduce__()


class stuf(directstuf, dict):

    '''dictionary with attribute-style access'''



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''setup stuf'''

import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

install_requires = ['distribute']
if sys.version_info[0] == 2 and sys.version_info[1] < 7:
    install_requires.extend(['ordereddict'])

setup(
    name='stuf',
    version='0.8.9',
    description='''stuf has attributes''',
    long_description=open(os.path.join(os.getcwd(), 'README.rst'), 'r').read(),
    author='L. C. Rees',
    url='https://bitbucket.org/lcrees/stuf/',
    author_email='lcrees@gmail.com',
    license='MIT',
    packages=['stuf'],
    test_suite='stuf.tests',
    zip_safe=False,
    keywords='dict attribute collection mapping dot notation access bunch',
    install_requires=install_requires,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
    ],
)
